const protobufjs = require('protobufjs');
const protobufCli = require('protobufjs/cli');
const ts = require('typescript');
const path = require('path');
const fs = require('fs');

const utils = require('./utils');

const getProtofiles = (protoDir, protoFile) => {
  return new Promise((resolve) => {
    protobufjs.load(path.resolve(protoDir, protoFile), (error, output) => {
      if (error) {
        console.error('getProtofiles [error]:', error.message);
        return [];
      }
      resolve(output.files);
    });
  });
};

const makeGeneratedJsonSchema = async (protoDir, protoFile, protoFiles) => {
  return new Promise((resolve, reject) => {
    protobufCli.pbjs.main(['--target', 'json', '--no-comments', ...protoFiles], (error, output) => {
      if (error) reject(error);

      const parsed = JSON.parse(output);
      const [protoShortname] = protoFile.split('.');
      const [packageName] = Object.keys(parsed.nested);

      const jsonSchema = {
        name: protoShortname,
        package: packageName,
        services: [],
      };

      const packageScheme = parsed.nested[packageName].nested;

      Object.keys(packageScheme).forEach((key) => {
        if (packageScheme[key].methods) {
          jsonSchema.services.push({ key, ...packageScheme[key] });
        }
      });

      resolve(jsonSchema);
    });
  });
};

const makeGeneratedProtoServices = async (protoDir, protoFile, protoFiles) => {
  return new Promise((resolve, reject) => {
    protobufCli.pbjs.main(
      [
        '--target',
        'static-module',
        '--wrap',
        'commonjs',
        '--sparse',
        '--no-create',
        '--no-verify',
        '--no-convert',
        '--no-delimited',
        ...protoFiles,
      ],
      (error, output) => {
        if (error) reject(error);
        resolve(output);
      },
    );
  });
};

const makeGeneratedProtoServicesDeclaration = async (name, servicesSourceFilename) => {
  return new Promise((resolve, reject) => {
    protobufCli.pbts.main([servicesSourceFilename], (error, output) => {
      if (error) reject(error);
      resolve(output);
    });
  });
};

const generateGrpcServiceFile = (jsonSchema, codegenDirectory) => {
  const createParameter = (name, type) => {
    return ts.createParameter(undefined, undefined, undefined, ts.createIdentifier(name), undefined, type);
  };
  const createArrowFunction = (name, parameters, block) => {
    return ts.createArrowFunction(
      name ? [ts.createIdentifier(`${name}`)] : undefined,
      undefined,
      parameters,
      undefined,
      undefined,
      ts.createBlock(block, true),
    );
  };
  const grpcWebNamedImport = ts.createImportDeclaration(
    undefined,
    undefined,
    ts.createNamedImports([
      ts.createIdentifier('AbstractClientBase'),
      ts.createIdentifier('GrpcWebClientBase'),
      ts.createIdentifier('Metadata'),
    ]),
    ts.createLiteral('grpc-web'),
  );
  const protoPackageImport = ts.createImportDeclaration(
    undefined,
    undefined,
    ts.createNamedImports([ts.createIdentifier(jsonSchema.package)]),
    ts.createLiteral(`./${codegenDirectory}/${jsonSchema.name}_pb`),
  );
  const extractMetadataFunction = ts.createArrowFunction(
    [ts.createIdentifier('private extractMetadata =')],
    undefined,
    [ts.createIdentifier('metadata: Metadata = {}')],
    undefined,
    undefined,
    ts.createBlock([ts.createIdentifier('return Object.assign({}, this.metadata, metadata);')], true),
  );
  const setMetadataFunction = ts.createArrowFunction(
    [ts.createIdentifier('setMetadata =')],
    undefined,
    [ts.createIdentifier('metadata: Metadata = {}')],
    undefined,
    undefined,
    ts.createBlock([ts.createIdentifier('this.metadata = Object.assign({}, this.metadata, metadata)')], true),
  );
  const makeServiceMethod = (packageName, service, method) => {
    const { requestType, responseType } = service.methods[method];
    return ts.createArrowFunction(
      [ts.createIdentifier(`${method}:`)],
      undefined,
      [ts.createIdentifier(`params: ${packageName}.I${requestType}`), ts.createIdentifier(`metadata?: Metadata`)],
      ts.createIdentifier(`Promise<${packageName}.${responseType}>`),
      undefined,
      ts.createBlock(
        [
          ts.createCall(ts.createIdentifier('return new Promise'), undefined, [
            createArrowFunction(
              undefined,
              [createParameter('resolve'), createParameter('reject')],
              [
                makeMethodInfo(packageName, requestType, responseType),
                makeClientRpcCallFunction(packageName, service.key, method),
              ],
            ),
          ]),
        ],
        true,
      ),
    );
  };
  const makeMethodInfo = (packageName, methodRequestType, methodResponseType) =>
    ts.createBinary(
      ts.createIdentifier('const methodInfo'),
      ts.createToken(ts.SyntaxKind.EqualsToken),
      ts.createNew(ts.createIdentifier('AbstractClientBase.MethodInfo'), undefined, [
        ts.createIdentifier(`${packageName}.${methodResponseType}`),
        createArrowFunction(
          undefined,
          [createParameter('request', ts.createIdentifier(`${packageName}.${methodRequestType}`))],
          [ts.createReturn(ts.createIdentifier(`${packageName}.${methodRequestType}.encode(request).finish()`))],
        ),
        ts.createIdentifier(`${packageName}.${methodResponseType}.decode`),
      ]),
    );
  const makeClientRpcCallFunction = (packageName, serviceName, methodName) =>
    ts.createCall(ts.createIdentifier('this.client.rpcCall'), undefined, [
      ts.createIdentifier(`this.hostname + '/${packageName}.${serviceName}/${methodName}'`),
      ts.createIdentifier('params'),
      ts.createIdentifier('this.extractMetadata(metadata)'),
      ts.createIdentifier('methodInfo'),
      createArrowFunction(
        undefined,
        [createParameter('error'), createParameter('response')],
        [ts.createIdentifier('if (error) reject(error);'), ts.createIdentifier('resolve(response);')],
      ),
    ]);
  const servicesStatement = jsonSchema.services.map((service) =>
    ts.createBinary(
      ts.createIdentifier(service.key),
      ts.SyntaxKind.EqualsToken,
      ts.createObjectLiteral(
        Object.keys(service.methods).map((method) => makeServiceMethod(jsonSchema.package, service, method)),
        true,
      ),
    ),
  );
  const grpcServiceClassDeclaration = ts.createClassDeclaration(
    undefined,
    [ts.createToken(ts.SyntaxKind.ExportKeyword)],
    ts.createIdentifier('GrpcService'),
    [],
    undefined,
    [
      createParameter('private client', ts.createIdentifier('GrpcWebClientBase')),
      createParameter(
        'private metadata',
        ts.createIdentifier("Metadata = { 'content-type': 'application/grpc-web' };"),
      ),
      createParameter('private hostname', ts.createToken(ts.SyntaxKind.StringKeyword)),
      ts.createConstructor(
        undefined,
        undefined,
        [createParameter('hostname', ts.createToken(ts.SyntaxKind.StringKeyword))],
        ts.createBlock(
          [
            ts.createBinary(
              ts.createIdentifier('this.client'),
              ts.SyntaxKind.EqualsToken,
              ts.createIdentifier('new GrpcWebClientBase({})'),
            ),
            ts.createBinary(
              ts.createIdentifier('this.hostname'),
              ts.SyntaxKind.EqualsToken,
              ts.createIdentifier('hostname'),
            ),
          ],
          true,
        ),
      ),
      extractMetadataFunction,
      setMetadataFunction,
      ...servicesStatement,
    ],
  );
  const sourceFile = ts.createSourceFile('someFileName.ts', '', ts.ScriptTarget.Latest, false, ts.ScriptKind.TS);
  const printer = ts.createPrinter({ newLine: ts.NewLineKind.Unspecified });
  const statements = [grpcWebNamedImport, protoPackageImport, grpcServiceClassDeclaration].map((statement) =>
    printer.printNode(ts.EmitHint.Unspecified, statement, sourceFile),
  );
  return statements.join('\n');
};

function disableEslint(source) {
  return ['/* eslint-disable */', source].join('\n');
}

(async () => {
  try {
    const moduleDir = path.resolve(__dirname, '..');

    const processArgv = {
      protoDir: utils.getArgValue(process.argv, '--proto-dir'),
      protoFile: utils.getArgValue(process.argv, '--proto-file'),
      outputDir: utils.getArgValue(process.argv, '--output-dir'),
      onlyProto: utils.getArgValue(process.argv, '--only-proto', JSON.parse),
    };

    if (!processArgv.protoDir || !processArgv.protoFile || !processArgv.outputDir) {
      process.exit(1);
    }

    const codegenDirectory = 'codegen';

    const outputProtoDir = path.resolve(processArgv.outputDir, codegenDirectory);
    fs.mkdirSync(outputProtoDir, { recursive: true });

    const protoFiles = await getProtofiles(processArgv.protoDir, processArgv.protoFile);

    const jsonSchema = await makeGeneratedJsonSchema(processArgv.protoDir, processArgv.protoFile, protoFiles);

    const generatedProtoSource = await makeGeneratedProtoServices(
      processArgv.protoDir,
      processArgv.protoFile,
      protoFiles,
    );
    const generatedProtoSourceFilename = path.resolve(outputProtoDir, `${jsonSchema.name}_pb.js`);

    fs.writeFileSync(generatedProtoSourceFilename, disableEslint(generatedProtoSource));

    const generatedProtoSourceDeclaration = await makeGeneratedProtoServicesDeclaration(
      jsonSchema.name,
      generatedProtoSourceFilename,
    );

    const generatedProtoSourceDeclarationFilename = path.resolve(outputProtoDir, `${jsonSchema.name}_pb.d.ts`);

    fs.writeFileSync(generatedProtoSourceDeclarationFilename, disableEslint(generatedProtoSourceDeclaration));

    if (!processArgv.onlyProto) {
      const grpcServicesSource = generateGrpcServiceFile(jsonSchema, codegenDirectory);
      const grpcServicesSourceFilename = path.resolve(processArgv.outputDir, 'GrpcService.ts');

      fs.writeFileSync(grpcServicesSourceFilename, disableEslint(grpcServicesSource));
    }
  } catch (error) {
    console.log(error);
  }
})();
